import time
import logging
import os
import subprocess
import yaml
import mmmanage.util as util
import mmmanage.monitor as monitor

def replication_state(replica=None):
    if replica is None:
        (exit, stdout, stderr) = util.local(["mysql", "-B", "-e", 'show slave status;'])
    else: 
        (exit, stdout, stderr) = util.ssh(replica, "mysql -B -e 'show slave status;'")
    if exit != 0:
        logging.error(f"Failed to get slave status from {replica}, exit code {exit}")
        print(stdout, stderr)

    result = stdout.decode("utf-8").split("\n")
    state = {}
    for k, v in zip(result[0].split("\t"), result[1].split("\t")):
        state[k] = v.replace("\\n", " ")
    return state


def read_global(ip, var):
    (exit, stdout, stderr) = util.ssh(ip, f"mysql -sre 'select @@global.{var}'")
    if exit != 0:
        logging.error(f"Failed to connect to {ip}, exit code {exit}")
    return stdout.decode("utf-8").replace("\n", "")


def read_config(file, server=None):
    if server is None:
        (exit, stdout, stderr) = util.local(["cat", file])
    else: 
        (exit, stdout, stderr) = util.ssh(server, f"cat {file}")
    if exit != 0:
        logging.error(f"Failed to connect to {server}, exit code {exit}")
    config = yaml.safe_load(stdout)
    return config


def set_read_only(check, ip, new_read_only, config_file):
    cur_read_only = read_global(ip, "read_only")
    expected_read_only = "0" if new_read_only else "1"
    if cur_read_only != expected_read_only:
        if new_read_only:
            raise Exception(f"MySQL server {ip} is already read-only, please investigate, are you trying to demote a replica?")
        else:
            raise Exception(f"MySQL server {ip} is not already read-only, please investigate")

    if check:
        logging.info(f"Would set mysql server {ip} read_only={new_read_only}")
    else:
        logging.info(f"Setting mysql server {ip} read_only={new_read_only}")
        # This config should already be written, but verify.
        (exit, stdout, stderr) = util.ssh(ip, f"grep -q read_only={new_read_only} {config_file}")
        if exit != 0:
            raise Exception(f"Could not find read_only={new_read_only} in {config_file} on {ip}")

        (exit, stdout, stderr) = util.ssh(ip, f"mysql -e 'set @@global.read_only={new_read_only};'")
        if exit != 0:
            raise Exception(f"Failed to set old read_only={new_read_only} on {ip}")

    if new_read_only == 1:
        (exit, stdout, stderr) = util.ssh(ip, f"mysql mysql -t -e 'select concat(user, \"@\", host) as user from user where super_priv=\"y\";'")
        t = str(stdout, 'UTF-8')
        logging.info(f"Current users have super_priv, so can still write to this replica:\n{t}")

def nftables_set_allowed_ips(ip, allowed_networks):
    allowed_networks_formatted = ", ".join(allowed_networks)
    logging.info(f"Updating nftables ruleset to only allow {allowed_networks_formatted}")
    nft_define = "# managed by mariadb-manage\ndefine mysql_allowed_networks = { " + allowed_networks_formatted + " }\n"
    (exit, stdout, stderr) = util.ssh(ip, f"cat > /etc/nftables-mariadb-manage.conf", bytes(nft_define, 'utf-8'))
    if exit != 0:
        raise Exception(f"Failed to write to config {config_file} on {ip}")
    (exit, stdout, stderr) = util.ssh(ip, f"systemctl reload nftables")
    if exit != 0:
        raise Exception(f"Failed to reload nftables ruleset")

def set_primary_firewall(check, config, ip):
    # Add firewall permits
    if check:
        logging.info(f"Would set primary firewall rules")
    else:
        if "primary_networks" in config:
            if config.get("nftables", False):
                permitted_networks = config.get('primary_networks')
                nftables_set_allowed_ips(ip, permitted_networks)
            else:
                permitted_networks = config.get('primary_networks')
                for nw in permitted_networks:
                    # Remove secondary block if present
                    logging.info(f"Removing firewall block for {nw} on {ip}")
                    (exit, stdout, stderr) = util.ssh(ip, f"ip6tables -C INPUT -p tcp -s {nw} --dport 3306 -m comment --comment 'mysql secondary automanaged' -j REJECT")
                    if exit == 0:
                        (exit, stdout, stderr) = util.ssh(ip, f"ip6tables -D INPUT -p tcp -s {nw} --dport 3306 -m comment --comment 'mysql secondary automanaged' -j REJECT")
                        if exit != 0:
                            raise Exception(f"Failed to remove {nw} from iptables rules on {ip}")
                    # Permit now we're primary
                    logging.info(f"Permitting firewall for {nw}")
                    (exit, stdout, stderr) = util.ssh(ip, f"ip6tables -C INPUT -p tcp -s {nw} --dport 3306 -m comment --comment 'mysql primary automanaged' -j ACCEPT")
                    if exit != 0:
                        (exit, stdout, stderr) = util.ssh(ip, f"ip6tables -I INPUT -p tcp -s {nw} --dport 3306 -m comment --comment 'mysql primary automanaged' -j ACCEPT")
                        if exit != 0:
                            raise Exception(f"Failed to add {nw} to iptables rules on {ip}")
                # Remove explicit grant for old primary
                replicas = config.get('replicas')
                for r in replicas:
                    oldprimary = r["ip"]
                    logging.info(f"Removing allow for old primary {oldprimary}")
                    (exit, stdout, stderr) = util.ssh(ip, f"ip6tables -C INPUT -p tcp -s {oldprimary} --dport 3306 -m comment --comment 'mysql secondary automanaged' -j ACCEPT")
                    if exit == 0:
                        (exit, stdout, stderr) = util.ssh(ip, f"ip6tables -D INPUT -p tcp -s {oldprimary} --dport 3306 -m comment --comment 'mysql secondary automanaged' -j ACCEPT")
                        if exit != 0:
                            raise Exception(f"Failed to remove {oldprimary} from iptables rules on {ip}")

                (exit, stdout, stderr) = util.ssh(ip, f"netfilter-persistent save")
                if exit != 0:
                    raise Exception(f"Failed save firewall rules")


def set_secondary_firewall(check, config, ip, newprimary):
    # Add firewall permits
    if check:
        logging.info(f"Would set secondary firewall rules")
    else:
        # Remove firewall permits
        if "primary_networks" in config:
            if config.get("nftables", False):
                nftables_set_allowed_ips(ip, [newprimary])
            else:
                permitted_networks = config.get('primary_networks')
                for nw in permitted_networks:
                    logging.info(f"Removing firewall permits for {nw} in {ip}")
                    (exit, stdout, stderr) = util.ssh(ip, f"ip6tables -C INPUT -p tcp -s {nw} --dport 3306 -m comment --comment 'mysql primary automanaged' -j ACCEPT")
                    if exit == 0:
                        (exit, stdout, stderr) = util.ssh(ip, f"ip6tables -D INPUT -p tcp -s {nw} --dport 3306 -m comment --comment 'mysql primary automanaged' -j ACCEPT")
                        if exit != 0:
                            raise Exception(f"Failed to remove {nw} from iptables rules on {ip}")
                    # And explicitly reject now we're secondary.
                    logging.info(f"Rejecting connections from {nw}")
                    (exit, stdout, stderr) = util.ssh(ip, f"ip6tables -C INPUT -p tcp -s {nw} --dport 3306 -m comment --comment 'mysql secondary automanaged' -j REJECT")
                    if exit != 0:
                        (exit, stdout, stderr) = util.ssh(ip, f"ip6tables -I INPUT -p tcp -s {nw} --dport 3306 -m comment --comment 'mysql secondary automanaged' -j REJECT")
                        if exit != 0:
                            raise Exception(f"Failed to add {nw} from iptables rules on {ip}")
                # Explicitly allow return traffic from the new primary in case it falls in the same ranges.
                logging.info(f"Explicitly permitting new primary {newprimary}")
                (exit, stdout, stderr) = util.ssh(ip, f"ip6tables -C INPUT -p tcp -s {newprimary} --dport 3306 -m comment --comment 'mysql secondary automanaged' -j ACCEPT")
                if exit != 0:
                    (exit, stdout, stderr) = util.ssh(ip, f"ip6tables -I INPUT -p tcp -s {newprimary} --dport 3306 -m comment --comment 'mysql secondary automanaged' -j ACCEPT")
                    if exit != 0:
                        raise Exception(f"Failed to add {newprimary} from iptables rules on {ip}")
                (exit, stdout, stderr) = util.ssh(ip, f"netfilter-persistent save")
                if exit != 0:
                    raise Exception(f"Failed save firewall rules")


def demote_primary(check, ip, newprimary):
    if check:
        logging.info(f"Would demote primary {ip}...")
    else:
        logging.info(f"Demoting primary {ip}...")

    lconfig = read_config("/etc/mythic/mariadb.yaml",ip)
    config_file = lconfig.get('config_file')
    c = create_config(lconfig)

    if not check:
        (exit, stdout, stderr) = util.ssh(ip, f"cat > {config_file}", bytes(c, 'utf-8'))
        if exit != 0:
            raise Exception(f"Failed to write to config {config_file} on {ip}")

    set_read_only(check, ip, 1, config_file)
    set_secondary_firewall(check, lconfig, ip, newprimary)


def make_replica(config, check, ip, primary):
    if check:
        logging.info(f"Would switch primary to {primary} for {ip}...")
    else:
        logging.info(f"Switching primary to {primary} for {ip}...")

        replica_pass = config.get("replica_password")

        runcmd = f"mysql -e \"CHANGE MASTER TO MASTER_HOST='{primary}',MASTER_USER='repl',MASTER_PASSWORD='{replica_pass}',MASTER_PORT=3306,MASTER_CONNECT_RETRY=10,master_use_gtid=slave_pos;\""
        if 'db_version' in config and config.get('db_version') == "percona8":
            logging.info(f"Using percona config...")
            runcmd = f"mysql -e \"CHANGE MASTER TO MASTER_HOST='{primary}',MASTER_USER='repl',MASTER_PASSWORD='{replica_pass}',MASTER_PORT=3306,MASTER_CONNECT_RETRY=10,MASTER_AUTO_POSITION=1, GET_MASTER_PUBLIC_KEY=1;\""
        (exit, stdout, stderr) = util.ssh(ip, runcmd)

        #(exit, stdout, stderr) = util.ssh(ip, f"mysql -e \"CHANGE MASTER TO MASTER_HOST='{primary}',MASTER_USER='repl',MASTER_PASSWORD='{replica_pass}',MASTER_PORT=3306,MASTER_CONNECT_RETRY=10,master_use_gtid=slave_pos;\"")
        if exit != 0:
                raise Exception("Couldn't change master.")
            
        (exit, stdout, stderr) = util.ssh(ip, "mysql -e 'start slave;'")
        if exit != 0:
                raise Exception("Couldn't start slave")
        # give it a few seconds to settle.
        time.sleep(5)
        state = replication_state(ip)
        if 'Slave_IO_Running' in state:
            print("  Slave_IO_Running:", state["Slave_IO_Running"])
            print("  Slave_SQL_Running:", state["Slave_SQL_Running"])
            print("  Seconds_Behind_Master:", state["Seconds_Behind_Master"])
            if 'db_version' in config and config.get('db_version') == "percona8":
                print("  Retrieved_Gtid_Set:", state["Retrieved_Gtid_Set"])
                print("  Executed_Gtid_Set:", state["Executed_Gtid_Set"])
            else:
                print("  Gtid_IO_Pos:", state["Gtid_IO_Pos"])
        else:
            logging.info("  WARNING, no slave state, check status")


def promote_secondary(check, config, ip, oldprimary):
    if check:
        logging.info(f"Would promote {ip} to primary DB server")

    if check:
        logging.info("Would stop mysql replication")
    else:
        logging.info("Stopping mysql replication")
        resetcmd = f"RESET MASTER;SET GLOBAL gtid_slave_pos = \"\";"
        stopcmd =  f"mysql -e 'stop all slaves; reset slave all;'"
        if 'db_version' in config and config.get('db_version') == "percona8":
            resetcmd = f"RESET MASTER;"
            stopcmd =  f"mysql -e 'stop slave; reset slave all;'"

        # We have to reset gtid_slave_pos here or future failovers will not work.
        (exit, stdout, stderr) = util.ssh(ip, stopcmd)
        if exit != 0:
            raise Exception(f"Failed to stop replication on {ip}")
        (exit, stdout, stderr) = util.local(["mysql", "-e", resetcmd])
        if exit != 0:
                raise Exception("Couldn't reset master/gtid replica pos.")

    lconfig = read_config("/etc/mythic/mariadb.yaml",ip)
    config_file = lconfig.get('config_file')
    c = create_config(lconfig, True)
    if check:
        logging.info(f"Would write primary config to {config_file}")
    else:
        logging.info(f"Writing primary config to {config_file}")
        (exit, stdout, stderr) = util.ssh(ip, f"cat > {config_file}", bytes(c, 'utf-8'))
        if exit != 0:
            raise Exception(f"Failed to write to config {config_file} on {ip}")
        print(config_file, c)
        logging.info(f"Marking primary read-write")

    set_read_only(check, ip, 0, config_file)
    set_primary_firewall(check, lconfig, ip)


def wait_for_repl(check, config, old, new):
    if check:
        logging.info("Would wait for replica to catch up")
        return

    # Completely different depending on version, this is percona8
    if 'db_version' in config and config.get('db_version') == "percona8":
        old_pos = read_global(old, "gtid_executed")
        logging.info(f"Old GTID position was {old_pos}, waiting for replica to catch up...")
        up_to_date = False
        while not up_to_date:
            c = f"select gtid_subset('{old_pos}', @@global.gtid_executed);"
            (exit, stdout, stderr) = util.local(["mysql", "-s", "-e", c])
            if exit != 0:
                raise Exception("Failed to gtid_subset! Is this mariadb?", stdout, stderr)
            result = stdout.decode("utf-8").split("\n")
            if result[0] == "1":
                logging.info("Replication is up to date")
                up_to_date = True
            else:
                logging.info(f"Old server was at {old_pos}, waiting 1s to check again...")
                time.sleep(1)

    # and mariadb 10.x
    else:
        old_pos = read_global(old, "gtid_current_pos")
        logging.info(f"Old GTID position was {old_pos}, waiting for replica to catch up...")

        # This is MariaDB specific, but works reliably.
        c = f"select master_gtid_wait('{old_pos}');"
        logging.info(f"Wait for replication: {c}")
        if not check:
            (exit, stdout, stderr) = util.local(["mysql", "-e", c])
            if exit != 0:
                raise Exception("Failed to master_gtid_wait! Is this mysql?", stdout, stderr)
            logging.info("Replication is up to date")


def grant_mysql(ip, replica_pass):
    logging.info(f"Setting replication grants for {ip}...")
    (exit, stdout, stderr) = util.local(["mysql", "-e", f"CREATE USER IF NOT EXISTS 'repl'@'{ip}' IDENTIFIED BY '{replica_pass}';"])
    if exit != 0:
        raise Exception("Failed to create user!")

    (exit, stdout, stderr) = util.local(["mysql", "-e", f"GRANT REPLICATION SLAVE ON *.* TO 'repl'@'{ip}';"])
    if exit != 0:
        raise Exception("Failed to grant replication permissions!")


def create_config(config, primary=False):
    serverid = config.get('me')['serverid']
    content = ["# *** THIS FILE IS AUTOMATICALLY MANAGED ***", "# Any changes will be overwritten!", "[mysqld]", "log-bin=mysql-bin", "expire_logs_days=2", f"server-id={serverid}","bind-address=::"]
    if 'db_version' in config and config.get('db_version') == "percona8":
        content.append("gtid_mode=ON")
        content.append("enforce-gtid-consistency=ON")
    else:
        content.append("gtid_strict_mode=1")
    # ,f"gtid_domain_id={serverid}"
    if primary:
        content.append("read_only=0")
        return "\n".join(content)
    else:
        content.append("read_only=1")
        return "\n".join(content)


def config_primary(config, check):
    # Write master config and restart mariadb
    config_file = config.get('config_file')

    if check:
        logging.info(f"Would write primary config to {config_file}")
    else:
        logging.info(f"Write primary config to {config_file}")
        
        f = open(config_file, "w")
        f.write(create_config(config, True))
        f.close()

    me = config.get('me')
    if check:
        logging.info(f"Would update mysql grants for replication")
    else:
        replica_pass = config.get("replica_password")
        grant_mysql(me["ip"], replica_pass)
        monitor.grant_monitor(config,me["ip"])
    
        replicas = config.get('replicas')
        for r in replicas:
            logging.info("replica: " + r['name'])
            grant_mysql(r["ip"], replica_pass)
            monitor.grant_monitor(config,r["ip"])

    dbservice = "mariadb"
    if 'db_version' in config and config.get('db_version') == "percona8":
        dbservice = "mysql"

    if check:
        logging.info(f"Would restart {dbservice}.")
    else:
        logging.info(f"Restarting {dbservice}...")

        (exit, stdout, stderr) = util.local(["systemctl","restart",dbservice])
        if exit != 0:
            raise Exception(f"Error restarting {dbservice}!")

    set_primary_firewall(check, config, me["ip"])


def clone_primary(config, ip, check):
    config_file = config.get('config_file')

    dbprocess = "mariadbd"
    dbservice = "mariadb"
    dbaptname = "mariadb-server"
    if 'db_version' in config and config.get('db_version') == "percona8":
        dbprocess = "mysqld"
        dbaptname = "percona-server-server"
        dbservice = "mysql"

    # Is mariadb currently running?
    (exit, stdout, stderr) = util.local(["pgrep", dbprocess])
    if (exit == 0): 
        raise Exception(f"{dbprocess} is running, aborting.")

    replica_pass = config.get("replica_password")
    (exit, stdout, stderr) = util.local(["mysql","-u","repl","-h",ip, f"--password={replica_pass}","-e","select 1;"])
    # exit code 0 means we connected OK...
    if exit != 0:
            raise Exception("Can't connect to primary!")

    # Ensure secondary firewall is configured
    me = config.get('me')["ip"]
    set_secondary_firewall(check, config, me, ip)

    if check:
        logging.info(f"Would purge {dbaptname}.")
    else:
        logging.info(f"Purging {dbaptname}.")
        # PURGE DOESN'T WHEN NONINTERACTIVE, WTF.
        # We'll remove the DB manually.
        (exit, stdout, stderr) = util.local(['rm','-rf','/var/lib/mysql'])
        if exit != 0:
            raise Exception("Error removing /var/lib/mysql")
        os.environ["DEBIAN_FRONTEND"] = "noninteractive"
        (exit, stdout, stderr) = util.local(['apt-get','-yqq','purge',f"{dbaptname}*"])
        if exit != 0:
            raise Exception(f"Error removing {dbaptname}")

    if check:
        logging.info(f"Would install {dbaptname}.")
    else:
        logging.info(f"Installing {dbaptname}.")
        (exit, stdout, stderr) = util.local(['apt-get','-y','install',dbaptname])
        if exit != 0:
            raise Exception(f"Error installing {dbaptname}")

    if check:
        logging.info(f"Would write replica config to {config_file}.")
    else:
        logging.info(f"Writing replica config to {config_file}.")
        f = open(config_file, "w")
        f.write(create_config(config))
        f.close()

    if check:
        logging.info(f"Would restart {dbservice}.")
    else:
        logging.info("Restarting mariadb...")
        (exit, stdout, stderr) = util.local(["systemctl","restart",dbservice])
        if exit != 0:
            raise Exception(f"Error restarting {dbservice}!")

    if check:
        logging.info("Would copy DB from primary.")
    else:
        logging.info("Copying DB from primary...")
        # Replicate DB - we probably want to use mariabackup in future or for bigger DBs.
        runcmd = ['ssh',f"root@{ip}", 'mysqldump', '--all-databases','--flush-privileges','--gtid','--master-data=1','--single-transaction']
        if 'db_version' in config and config.get('db_version') == "percona8":
            runcmd = ['ssh',f"root@{ip}", 'mysqldump', '--all-databases','--flush-privileges','--source-data=1','--single-transaction']
        dump = subprocess.Popen(runcmd, stdout=subprocess.PIPE)
        my = subprocess.Popen('mysql', stdin=dump.stdout,stdout=subprocess.PIPE)
        my.communicate()

    if check:
        logging.info("Would start replication.")
    else:
        logging.info(f"Starting replication...")

        if 'db_version' in config and config.get('db_version') != "percona8":
            # We need this for mariadb, but it breaks percona
            (exit, stdout, stderr) = util.local(["mysql", "-e", f"RESET MASTER;"])
            if exit != 0:
                    raise Exception("Couldn't reset master.")
        runcmd = ["mysql", "-e", f"CHANGE MASTER TO MASTER_HOST='{ip}',MASTER_USER='repl',MASTER_PASSWORD='{replica_pass}',MASTER_PORT=3306,MASTER_CONNECT_RETRY=10,master_use_gtid=slave_pos;"]
        if 'db_version' in config and config.get('db_version') == "percona8":
            logging.info(f"Using percona config...")
            runcmd = ["mysql", "-e", f"CHANGE MASTER TO MASTER_HOST='{ip}',MASTER_USER='repl',MASTER_PASSWORD='{replica_pass}',MASTER_PORT=3306,MASTER_CONNECT_RETRY=10,MASTER_AUTO_POSITION=1, GET_MASTER_PUBLIC_KEY=1;"]
        (exit, stdout, stderr) = util.local(runcmd)
        if exit != 0:
                raise Exception("Couldn't change master.")
        (exit, stdout, stderr) = util.local(["mysql", "-e", "start slave;"])
        if exit != 0:
                raise Exception("Couldn't start slave")

        state = replication_state()
        if 'Slave_IO_Running' in state:
            print("  Slave_IO_Running:", state["Slave_IO_Running"])
            print("  Slave_SQL_Running:", state["Slave_SQL_Running"])
            print("  Seconds_Behind_Master:", state["Seconds_Behind_Master"])
            if 'db_version' in config and config.get('db_version') == "percona8":
                print("  Retrieved_Gtid_Set:", state["Retrieved_Gtid_Set"])
                print("  Executed_Gtid_Set:", state["Executed_Gtid_Set"])
            else:
                print("  Gtid_IO_Pos:", state["Gtid_IO_Pos"])
        else:
            print("  WARNING, no slave state!")
